from enum import Enum
from threading import Thread
from typing import Dict, Callable, Optional

import keyboard
from playsound import playsound

from fishy.helper import helper


class Key(Enum):
    F9 = "f9"
    F10 = "f10"
    F8 = "f8"
    F7 = "f7"
    UP = "up"
    DOWN = "down"
    LEFT = "left"
    RIGHT = "right"


_hotkeys: Dict[Key, Optional[Callable]] = {}


def _get_callback(k):
    def callback():
        if not _hotkeys[k]:
            return

        playsound(helper.manifest_file("beep.wav"), False)
        Thread(target=_hotkeys[k]).start()
    return callback


def initalize():
    for k in Key:
        _hotkeys[k] = None
        keyboard.add_hotkey(k.value, _get_callback(k))


def set_hotkey(key: Key, func: Optional[Callable]):
    _hotkeys[key] = func


def free_key(k: Key):
    set_hotkey(k, None)



import math
import logging
import time

import cv2
import numpy as np

from fishy.engine.fullautofisher.engine import FullAuto
from pynput import keyboard, mouse

from fishy.helper import hotkey
from fishy.helper.config import config
from fishy.helper.helper import wait_until
from fishy.helper.hotkey import Key

mse = mouse.Controller()
kb = keyboard.Controller()

offset = 0


def get_crop_coods(window):
    img = window.get_capture()
    img = cv2.inRange(img, 0, 1)

    cnt, h = cv2.findContours(img, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

    """
    code from https://stackoverflow.com/a/45770227/4512396
    """
    for i in range(len(cnt)):
        area = cv2.contourArea(cnt[i])
        if 5000 < area < 100000:
            mask = np.zeros_like(img)
            cv2.drawContours(mask, cnt, i, 255, -1)
            x, y, w, h = cv2.boundingRect(cnt[i])
            return x, y + offset, x + w, y + h - offset


def _update_factor(key, value):
    full_auto_factors = config.get("full_auto_factors", {})
    full_auto_factors[key] = value
    config.set("full_auto_factors", full_auto_factors)


def _get_factor(key):
    return config.get("full_auto_factors", {}).get(key)


class Calibrate:
    def __init__(self, engine: FullAuto):
        self._callibrate_state = -1
        self.engine = engine

    # region getters
    @property
    def crop(self):
        return _get_factor("crop")

    @property
    def move_factor(self):
        return _get_factor("move_factor")

    @property
    def rot_factor(self):
        return _get_factor("rot_factor")

    @property
    def time_to_reach_bottom(self):
        return _get_factor("time_to_reach_bottom")

    # endregion

    def all_callibrated(self):
        return self.crop is not None and self.move_factor is not None and self.rot_factor is not None

    def toggle_show(self):
        self.engine.show_crop = not self.engine.show_crop

    def update_crop(self, enable_crop=True):
        if enable_crop:
            self.engine.show_crop = True
        crop = get_crop_coods(self.engine.window)
        _update_factor("crop", crop)
        self.engine.window.crop = crop

    def walk_calibrate(self):
        walking_time = 3

        coods = self.engine.get_coods()
        if coods is None:
            return

        x1, y1, rot1 = coods

        kb.press('w')
        time.sleep(walking_time)
        kb.release('w')
        time.sleep(0.5)

        coods = self.engine.get_coods()
        if coods is None:
            return
        x2, y2, rot2 = coods

        move_factor = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2) / walking_time
        _update_factor("move_factor", move_factor)
        logging.info("done")

    def rotate_calibrate(self):
        rotate_times = 50

        coods = self.engine.get_coods()
        if coods is None:
            return
        _, _, rot2 = coods

        for _ in range(rotate_times):
            mse.move(FullAuto.rotate_by, 0)
            time.sleep(0.05)

        coods = self.engine.get_coods()
        if coods is None:
            return
        x3, y3, rot3 = coods

        if rot3 > rot2:
            rot3 -= 360

        rot_factor = (rot3 - rot2) / rotate_times
        _update_factor("rot_factor", rot_factor)
        logging.info("done")

    def time_to_reach_bottom_callibrate(self):
        self._callibrate_state = 0

        def _f8_pressed():
            self._callibrate_state += 1

        logging.info("look straight up and press f8")
        hotkey.set_hotkey(Key.F8, _f8_pressed)

        wait_until(lambda: self._callibrate_state == 1)

        logging.info("as soon as you look on the floor, press f8 again")

        y_cal_start_time = time.time()
        while self._callibrate_state == 1:
            mse.move(0, FullAuto.rotate_by)
            time.sleep(0.05)
        hotkey.free_key(Key.F8)

        time_to_reach_bottom = time.time() - y_cal_start_time
        _update_factor("time_to_reach_bottom", time_to_reach_bottom)
        logging.info("done")



import math
import os
import tempfile
import traceback
import uuid
from enum import Enum
from threading import Thread
from zipfile import ZipFile

import cv2
import logging
import time

import numpy as np
import pytesseract

from fishy.engine.fullautofisher.tesseract import is_tesseract_installed, downlaoad_and_extract_tesseract, \
    get_values_from_image
from fishy.engine.semifisher.fishing_mode import FishingMode

from fishy.engine import SemiFisherEngine
from fishy.engine.common.window import WindowClient
from fishy.engine.semifisher import fishing_mode, fishing_event

from fishy.engine.common.IEngine import IEngine
from pynput import keyboard, mouse

from fishy.helper import hotkey, helper
from fishy.helper.config import config
from fishy.helper.downloader import download_file_from_google_drive
from fishy.helper.helper import sign
from fishy.helper.hotkey import Key

mse = mouse.Controller()
kb = keyboard.Controller()


def image_pre_process(img):
    scale_percent = 200  # percent of original size
    width = int(img.shape[1] * scale_percent / 100)
    height = int(img.shape[0] * scale_percent / 100)
    dim = (width, height)
    img = cv2.resize(img, dim, interpolation=cv2.INTER_AREA)
    img = cv2.bitwise_not(img)
    return img


class State(Enum):
    NONE = 0
    PLAYING = 1
    RECORDING = 2
    OTHER = 3


class FullAuto(IEngine):
    rotate_by = 30
    state = State.NONE

    def __init__(self, gui_ref):
        from fishy.engine.fullautofisher.controls import Controls
        from fishy.engine.fullautofisher import controls
        from fishy.engine.fullautofisher.calibrate import Calibrate
        from fishy.engine.fullautofisher.test import Test

        super().__init__(gui_ref)
        self._hole_found_flag = False
        self._curr_rotate_y = 0

        self.fisher = SemiFisherEngine(None)
        self.calibrate = Calibrate(self)
        self.test = Test(self)
        self.controls = Controls(controls.get_controls(self))
        self.show_crop = False

    def run(self):
        self.show_crop = False
        FullAuto.state = State.NONE

        self.gui.bot_started(True)
        fishing_event.unsubscribe()
        self.fisher.toggle_start()

        self.window = WindowClient(color=cv2.COLOR_RGB2GRAY, show_name="Full auto debug")

        try:
            if self.calibrate.crop is None:
                self.calibrate.update_crop(enable_crop=False)
            self.window.crop = self.calibrate.crop

            if not is_tesseract_installed():
                logging.info("tesseract not found")
                downlaoad_and_extract_tesseract()

            if not self.calibrate.all_callibrated():
                logging.error("you need to callibrate first")

            self.controls.initialize()
            while self.start and WindowClient.running():
                self.window.show(self.show_crop, func=image_pre_process)
                if not self.show_crop:
                    time.sleep(0.1)
        except:
            traceback.print_exc()

            if self.window.get_capture() is None:
                logging.error("Game window not found")

        self.gui.bot_started(False)
        self.controls.unassign_keys()
        self.window.show(False)
        logging.info("Quitting")
        self.window.destory()
        self.fisher.toggle_start()

    def get_coods(self):
        img = self.window.processed_image(func=image_pre_process)
        return get_values_from_image(img)

    def move_to(self, target):
        if target is None:
            logging.error("set target first")
            return

        if not self.calibrate.all_callibrated():
            logging.error("you need to callibrate first")
            return

        current = self.get_coods()
        print(f"Moving from {(current[0], current[1])} to {target}")
        move_vec = target[0] - current[0], target[1] - current[1]

        dist = math.sqrt(move_vec[0] ** 2 + move_vec[1] ** 2)
        print(f"distance: {dist}")
        if dist < 5e-05:
            print("distance very small skipping")
            return

        target_angle = math.degrees(math.atan2(-move_vec[1], move_vec[0])) + 90
        from_angle = current[2]

        self.rotate_to(target_angle, from_angle)

        walking_time = dist / self.calibrate.move_factor
        print(f"walking for {walking_time}")
        kb.press('w')
        time.sleep(walking_time)
        kb.release('w')
        print("done")

    def rotate_to(self, target_angle, from_angle=None):
        if from_angle is None:
            _, _, from_angle = self.get_coods()

        if target_angle < 0:
            target_angle = 360 + target_angle
        while target_angle > 360:
            target_angle -= 360
        print(f"Rotating from {from_angle} to {target_angle}")

        angle_diff = target_angle - from_angle

        if abs(angle_diff) > 180:
            angle_diff = (360 - abs(angle_diff)) * sign(angle_diff) * -1

        rotate_times = int(angle_diff / self.calibrate.rot_factor) * -1

        print(f"rotate_times: {rotate_times}")

        for _ in range(abs(rotate_times)):
            mse.move(sign(rotate_times) * FullAuto.rotate_by * -1, 0)
            time.sleep(0.05)

    def look_for_hole(self):
        self._hole_found_flag = False

        if FishingMode.CurrentMode == fishing_mode.State.LOOK:
            return True

        def found_hole(e):
            if e == fishing_mode.State.LOOK:
                self._hole_found_flag = True

        fishing_mode.subscribers.append(found_hole)

        t = 0
        while not self._hole_found_flag and t <= self.calibrate.time_to_reach_bottom / 3:
            mse.move(0, FullAuto.rotate_by)
            time.sleep(0.05)
            t += 0.05
        while not self._hole_found_flag and t > 0:
            mse.move(0, -FullAuto.rotate_by)
            time.sleep(0.05)
            t -= 0.05

        self._curr_rotate_y = t
        fishing_mode.subscribers.remove(found_hole)
        return self._hole_found_flag

    def rotate_back(self):
        while self._curr_rotate_y > 0.01:
            mse.move(0, -FullAuto.rotate_by)
            time.sleep(0.05)
            self._curr_rotate_y -= 0.05

    def toggle_start(self):
        if self.start and FullAuto.state != State.NONE:
            logging.info("Please turn off RECORDING/PLAYING first")
            return

        self.start = not self.start
        if self.start:
            self.thread = Thread(target=self.run)
            self.thread.start()


if __name__ == '__main__':
    logging.getLogger("").setLevel(logging.DEBUG)
    hotkey.initalize()
    # noinspection PyTypeChecker
    bot = FullAuto(None)
    bot.toggle_start()



import logging
import os
import tempfile
import uuid
from datetime import datetime
from zipfile import ZipFile

import cv2

import pytesseract

from fishy.helper.downloader import download_file_from_google_drive
from fishy.helper.helper import get_documents

directory = os.path.join(os.environ["APPDATA"], "Tesseract-OCR")


def downlaoad_and_extract_tesseract():
    logging.info("Tesseract-OCR downlaoding, Please wait...")

    f = tempfile.NamedTemporaryFile(delete=False)
    download_file_from_google_drive("16llzcBlaCsG9fm-rY2dD4Gvopnhm3XoE", f)
    f.close()

    logging.info("Tesseract-OCR downloaded, now installing")

    with ZipFile(f.name, 'r') as z:
        z.extractall(path=directory)

    logging.info("Tesseract-OCR installed")


def is_tesseract_installed():
    return os.path.exists(os.path.join(os.environ["APPDATA"], "Tesseract-OCR"))


# noinspection PyBroadException
def get_values_from_image(img):
    try:
        pytesseract.pytesseract.tesseract_cmd = directory + '/tesseract.exe'
        tessdata_dir_config = f'--tessdata-dir "{directory}" -c tessedit_char_whitelist=0123456789.'

        text = pytesseract.image_to_string(img, lang="eng", config=tessdata_dir_config)
        text = text.replace(" ", "")
        vals = text.split(":")
        return float(vals[0]), float(vals[1]), float(vals[2])
    except Exception:
        logging.error("Couldn't read coods, make sure 'crop' calibration is correct")
        cv2.imwrite(os.path.join(get_documents(), "fishy_failed_reads", f"{datetime.now()}.jpg"), img)
        return None



import time
from tkinter import *
from tkinter import messagebox
from tkinter.ttk import *

import typing

from fishy.helper import helper

from fishy.web import web

from fishy.libs.tkhtmlview import HTMLLabel
from ..helper.config import config

if typing.TYPE_CHECKING:
    from . import GUI


# noinspection PyProtectedMember
def discord_login(gui: 'GUI'):
    if web.is_logged_in(config.get("uid")):
        if web.logout(config.get("uid")):
            gui.login.set(0)
        return

    # set notification checkbutton
    gui.login.set(0)

    def quit_top():
        top.destroy()
        top_running[0] = False

    def check():
        code = int(login_code.get()) if login_code.get().isdigit() else 0
        if web.login(config.get("uid"), code):
            gui.login.set(1)
            messagebox.showinfo("Note!", "Logged in successfuly!")
            quit_top()
        else:
            messagebox.showerror("Error", "Logged wasn't successful")

    top_running = [True]

    top = Toplevel(background=gui._root["background"])
    top.minsize(width=300, height=300)
    top.title("Notification Setup")

    html_label = HTMLLabel(top,
                           html=f'<div style="color: {gui._console["fg"]}; text-align: center">'
                                f'<p><span style="font-size:20px">Step 1.</span><br/>'
                                f'Join <a href="https://discord.definex.in/">Discord server</a></p>'
                                f'<p><span style="font-size:20px">Step 2.</span><br/>'
                                f'run !login command in #bot-spam channel'
                                f'<p><span style="font-size:20px">Step 3.</span><br/>'
                                f'enter login code'
                                f'</div>', background=gui._root["background"])

    html_label.pack(pady=(20, 5))
    html_label.fit_height()

    login_code = Entry(top, justify=CENTER, font="Calibri 15")
    login_code.pack(padx=(15, 15), expand=True, fill=BOTH)

    html_label = HTMLLabel(top,
                           html=f'<div style="color: {gui._console["fg"]}; text-align: center">'
                                f'<p><span style="font-size:20px">Step 4.</span><br/></p>'
                                f'</div>', background=gui._root["background"])

    html_label.pack(pady=(5, 5))
    html_label.fit_height()

    Button(top, text="REGISTER", command=check).pack(pady=(5, 20))

    top.protocol("WM_DELETE_WINDOW", quit_top)
    top.grab_set()
    while top_running[0]:
        top.update()
        time.sleep(0.01)
    top.grab_release()



import logging

from fishy.helper import hotkey, helper

from fishy.engine.fullautofisher.engine import FullAuto, State
from fishy.helper.hotkey import Key


def get_controls(engine: FullAuto):
    from fishy.engine.fullautofisher.recorder import Recorder
    from fishy.engine.fullautofisher.player import Player
    controls = [
        ("MODE_SELECT", {
            Key.RIGHT: (lambda: engine.controls.select_mode("CALIBRATE"), "calibrate mode"),
            Key.UP: (lambda: engine.controls.select_mode("TEST1"), "test mode"),
            Key.LEFT: (Player(engine).toggle_move, "start/stop play"),
            Key.DOWN: (Recorder(engine).toggle_recording, "start/stop record"),
        }),
        ("CALIBRATE", {
            Key.RIGHT: (engine.calibrate.update_crop, "cropping"),
            Key.UP: (engine.calibrate.walk_calibrate, "walking"),
            Key.LEFT: (engine.calibrate.rotate_calibrate, "rotation"),
            Key.DOWN: (engine.calibrate.time_to_reach_bottom_callibrate, "look up down")
        }),
        ("TEST1", {
            Key.RIGHT: (engine.test.print_coods, "print coordinates"),
            Key.UP: (engine.test.look_for_hole, "look for hole up down"),
            Key.LEFT: (engine.test.set_target, "set target"),
            Key.DOWN: (engine.test.move_to_target, "move to target")
        })
    ]

    return controls


class Controls:
    def __init__(self, controls, first=0):
        self.current_menu = first - 1
        self.controls = controls

    def initialize(self):
        self.select_mode(self.controls[0][0])

    def log_help(self):
        help_str = f"\nCONTROLS: {self.controls[self.current_menu][0]}"
        for key, meta in self.controls[self.current_menu][1].items():
            func, name = meta
            if func:
                hotkey.set_hotkey(key, func)
                help_str += f"\n{key.value}: {name}"
        logging.info(help_str)

    def select_mode(self, mode):
        if FullAuto.state != State.NONE:
            self.log_help()
            return

        self.current_menu = 0
        for i, control in enumerate(self.controls):
            if mode == control[0]:
                self.current_menu = i
        self.log_help()

    def unassign_keys(self):
        keys = []
        for c in self.controls:
            for k in c[1].keys():
                if k not in keys:
                    hotkey.free_key(k)



"""
fishing_event.py
Defines different fishing modes (states) which acts as state for state machine
also implements callbacks which is called when states are changed
"""
import logging
import time

from fishy.engine.semifisher import fishing_mode
from playsound import playsound

from fishy import web
from fishy.engine.semifisher.fishing_mode import State, FishingMode
from fishy.helper import helper
import keyboard

from fishy.helper.config import config


class FishEvent:
    fishCaught = 0
    totalFishCaught = 0
    stickInitTime = 0
    fish_times = []
    hole_start_time = 0
    FishingStarted = False
    previousState = State.IDLE

    # initialize these
    action_key = 'e'
    collect_r = False
    uid = None
    sound = False


def init():
    subscribe()
    FishEvent.action_key = config.get("action_key", 'e')
    FishEvent.uid = config.get("uid")
    FishEvent.sound = config.get("sound_notification", False)


def unsubscribe():
    if fisher_callback in fishing_mode.subscribers:
        fishing_mode.subscribers.remove(fisher_callback)


def subscribe():
    if fisher_callback not in fishing_mode.subscribers:
        fishing_mode.subscribers.append(fisher_callback)

        if FishingMode.CurrentMode == State.LOOK:
            fisher_callback(FishingMode.CurrentMode)


def fisher_callback(event: State):
    callbacks_map = {State.HOOK: on_hook, State.LOOK: on_look, State.IDLE: on_idle, State.STICK: on_stick}
    callbacks_map[event]()
    FishEvent.previousState = event


def on_hook():
    """
    called when the fish hook is detected
    increases the `fishCaught`  and `totalFishCaught`, calculates the time it took to catch
    presses e to catch the fish
    """
    FishEvent.fishCaught += 1
    FishEvent.totalFishCaught += 1
    time_to_hook = time.time() - FishEvent.stickInitTime
    FishEvent.fish_times.append(time_to_hook)
    logging.info("HOOOOOOOOOOOOOOOOOOOOOOOK....... " + str(FishEvent.fishCaught) + " caught " + "in " + str(
        round(time_to_hook, 2)) + " secs.  " + "Total: " + str(FishEvent.totalFishCaught))

    keyboard.press_and_release(FishEvent.action_key)

    if FishEvent.collect_r:
        time.sleep(0.1)
        keyboard.press_and_release('r')
        time.sleep(0.1)


def on_look():
    """
    presses e to throw the fishing rod
    """
    keyboard.press_and_release(FishEvent.action_key)


def on_idle():
    if FishEvent.fishCaught > 0:
        web.send_hole_deplete(FishEvent.uid, FishEvent.fishCaught, time.time() - FishEvent.hole_start_time,
                              FishEvent.fish_times)
        FishEvent.fishCaught = 0

    if FishEvent.previousState == State.HOOK:
        logging.info("HOLE DEPLETED")
    else:
        logging.info("FISHING INTERRUPTED")

    if FishEvent.sound:
        playsound(helper.manifest_file("sound.mp3"), False)


def on_stick():
    FishEvent.stickInitTime = time.time()
    FishEvent.FishingStarted = True

    if FishEvent.fishCaught == 0:
        FishEvent.hole_start_time = time.time()
        FishEvent.fish_times = []



import logging
import pickle
import time
from pprint import pprint
from tkinter.filedialog import asksaveasfile

from fishy.engine.fullautofisher.engine import FullAuto, State

from fishy.helper import hotkey
from fishy.helper.hotkey import Key


class Recorder:
    recording_fps = 1
    mark_hole_key = Key.F8

    def __init__(self, engine: FullAuto):
        self.recording = False
        self.engine = engine
        self.timeline = []

    def _mark_hole(self):
        coods = self.engine.get_coods()
        self.timeline.append(("check_fish", coods))
        logging.info("check_fish")

    def toggle_recording(self):
        if FullAuto.state != State.RECORDING and FullAuto.state != State.NONE:
            return

        self.recording = not self.recording
        if self.recording:
            self._start_recording()

    def _start_recording(self):
        FullAuto.state = State.RECORDING
        logging.info("starting, press f8 to mark hole")
        hotkey.set_hotkey(Recorder.mark_hole_key, self._mark_hole)

        self.timeline = []

        while self.recording:
            start_time = time.time()
            coods = None
            while not coods:
                coods = self.engine.get_coods()
            self.timeline.append(("move_to", (coods[0], coods[1])))

            time_took = time.time() - start_time
            if time_took <= Recorder.recording_fps:
                time.sleep(Recorder.recording_fps - time_took)
            else:
                logging.warning("Took too much time to record")

        hotkey.free_key(Recorder.mark_hole_key)

        def func():
            _file = None
            files = [('Fishy File', '*.fishy')]
            while not _file:
                _file = asksaveasfile(mode='wb', filetypes=files, defaultextension=files)

            return _file

        file = self.engine.get_gui().call_in_thread(func, block=True)
        data = {"full_auto_path": self.timeline}
        pprint(data)
        pickle.dump(data, file)
        file.close()
        FullAuto.state = State.NONE




import requests


def download_file_from_google_drive(id, file):
    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params={'id': id}, stream=True)
    token = get_confirm_token(response)

    if token:
        params = {'id': id, 'confirm': token}
        response = session.get(URL, params=params, stream=True)

    save_response_content(response, file)


def get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None


def save_response_content(response, f):
    CHUNK_SIZE = 32768
    for chunk in response.iter_content(CHUNK_SIZE):
        if chunk:  # filter out keep-alive new chunks
            f.write(chunk)



import typing
from abc import ABC, abstractmethod
from threading import Thread
from typing import Callable

from fishy.gui.funcs import GUIFuncsMock

if typing.TYPE_CHECKING:
    from fishy.gui import GUI


class IEngine(ABC):

    def __init__(self, gui_ref: 'Callable[[], GUI]'):
        self.get_gui = gui_ref
        self.start = False
        self.window = None
        self.thread = None

    @property
    def gui(self):
        if self.get_gui is None:
            return GUIFuncsMock()

        return self.get_gui().funcs

    @abstractmethod
    def toggle_start(self):
        ...

    @abstractmethod
    def run(self):
        ...
